<div class="control-group" data-bind="visible: config_checkType() === 'github_commit'">
    <label class="control-label">{{ _('Tracked branch') }}</label>
    <div class="controls">
        <input type="text" class="input-block-level" data-bind="value: config_trackedBranch" placeholder="master">
    </div>
</div>
