#settings_plugin_backup_backup_table {
  .settings_plugin_backup_name {
  }
  .settings_plugin_backup_actions {
    width: 75px;
    text-align: center;

    a {
      color: black;
    }
  }
  .settings_plugin_backup_date {
    width: 200px;
    text-align: right;
  }
  .settings_plugin_backup_size {
    width: 100px;
    text-align: right;
  }

  .settings_plugin_backup_checkbox {
    text-align: center;
    width: 10px;

    input[type="checkbox"] {
      margin-top: 0;
    }
  }
}

#settings_plugin_backup_restoredialog_output {
  .message {
    font-weight: bold;
  }

  .error {
    font-weight: bold;
    color: #990000;
  }

  .stdout {
    color: #333333;
  }

  .stderr {
    color: #990000;
  }

  .call {
    color: #000099;
  }
}
