.achievement-list .achievement-item .achievement-logo svg,.achievement-popup .achievement .achievement-logo svg{fill:currentColor}.achievement-list{display:flex;width:100%;flex-wrap:wrap}.achievement-list .achievement-item{display:flex;flex:1 1 calc(50% - 2.5em);max-width:calc(50% - 2.5em);padding:.5em;margin:.5em;border:1px solid #666;border-radius:.5em;opacity:.5}.achievement-list .achievement-item.achieved{color:#fff;background-color:#08c;border:1px solid #069;opacity:1}.achievement-list .achievement-item .achievement-logo{width:48px;height:48px;min-width:48px;margin-right:.5em}.achievement-popup .achievement{display:inline-flex;margin-top:1em}.achievement-popup .achievement .achievement-logo{width:48px;height:48px;min-width:48px;margin-right:.5em}.achievement-popup .achievement .achievement-description{flex-grow:1}.achievement-popup .achievement-nag{margin-top:1em}.achievement-popup .achievement-link{display:block;margin-top:1em}