<h3>{{ _('All Done!') }}</h3>

<p>{% trans %}
    Your OctoPrint installation is now all set up and ready to go.
{% endtrans %}</p>

<h4>{{ _('Please read this for your and your printer\'s safety!') }}</h4>

<div class="row-fluid">
    <div class="span1 text-center">
        <i class="fas fa-fire-extinguisher fa-3x text-error"></i>
    </div>
    <div class="span11">
        <p>{% trans %}
            Even with OctoPrint you should <strong class="text-error">never leave
            your printer running unattended</strong>. The electronics in our consumer printers can and
            sometimes sadly do catch fire, so stay or have someone else stay close enough to physically intervene
            in case of catastrophic failure.
        {% endtrans %}</p>
        <p><label class="checkbox" data-bind="style: { border: !dontPrintUnattendedAcknowledged() && error() ? '1px dashed red' : 'inherit' }">
            <input type="checkbox" data-bind="checked: dontPrintUnattendedAcknowledged">
            <strong>{{ _('I have read and understood this') }}</strong><i class="fas fa-left-arrow text-error"></i>
        </label></p>
    </div>
</div>

<div class="row-fluid">
    <div class="span1 text-center">
        <i class="fas fa-unlink fa-3x text-error"></i>
    </div>
    <div class="span11">
        <p>{% trans url='https://plugins.octoprint.org/topics/remote_access/' %}
            You also should <strong class="text-error">never expose OctoPrint on the public internet</strong>,
            e.g. through a port forward. It can and <em>will</em> be found, and bad people
            <em>will</em> try to abuse it. Don't put yourself at risk, use a safe way to access your instance remotely, like
            one of the available <strong><a href="{{ url }}" target="_blank" rel="noopener noreferrer">Remote Access plugins!</a></strong>
        {% endtrans %}</p>
        <p><label class="checkbox" data-bind="style: { border: !dontPortForwardAcknowledged() && error() ? '1px dashed red' : 'inherit' }">
            <input type="checkbox" data-bind="checked: dontPortForwardAcknowledged">
            <strong>{{ _('I have read and understood this') }}</strong>
        </label></p>
    </div>
</div>

<h4>{{ _('Support OctoPrint\'s ongoing development!') }}</h4>

<div class="row-fluid">
    <div class="span1 text-center"><img src="{{ url_for('static', filename='img/tentacle-56x56.png') }}"></div>
    <div class="span11">
        <p>{% trans url='https://octoprint.org/support-octoprint/?utm_source=octoprint&utm_medium=first_run_wizard&campaign=funding' %}
            Work on OctoPrint is only possible through continued financial support by people like you!
            If you enjoy OctoPrint, please <strong><a href="{{ url }}" target="_blank">click here to learn how to help with its funding</a></strong> -
            every bit helps!<br>
            <strong>Thank you!</strong>
        {% endtrans %}</p>
        <p><label class="checkbox" data-bind="style: { border: !fundingRequestAcknowledged() && error() ? '1px dashed red' : 'inherit' }">
            <input type="checkbox" data-bind="checked: fundingRequestAcknowledged">
            <strong>{{ _('I have read and understood this') }}</strong>
        </label></p>
    </div>
</div>

<p>{% trans %}
    With all that being said, <strong>Happy and Safe Printing</strong>!
{% endtrans %}</p>
