<a href="https://faq.octoprint.org/plugintimings-log" target="_blank" title="{{ _('plugintimings.log is enabled') }}" aria-label="{{ _('plugintimings.log is enabled') }}" rel="noopener noreferrer" data-bind="visible: pluginTimingsLogEnabled(), popover: { placement: 'bottom', title: '{{ _('<code>plugintimings.log</code> is enabled')|esq }}', html: true, content: pluginTimingsLogPopoverContent, trigger: 'hover' }" style="display: none">
    <i class="fas fa-file-alt"></i>
</a>
