<div class="control-group" title="{{ _('Number of FFMPEG encoding threads')|edq }}">
    <label class="control-label" for="settings-webcamFfmpegThreads">{{ _('FFMPEG threads') }}</label>
    <div class="controls">
        <input class="input-mini" data-bind="value: webcam_ffmpegThreads" id="settings-webcamFfmpegThreads" type="number" step="1" min="1">
    </div>
</div>
