<div id="timelapsePreviewModal" class="modal large hide fade-in">
	<div class="modal-header">
		<a href="#" class="close" data-dismiss="modal" aria-hidden="true">&times;</a>
		<h3>{{ _('Timelapse Preview') }}</h3>
	</div>
	<div class="modal-body" style="text-align: center">
		<video width="800px" height="600px" style="width: 800px; height: 600px" autoplay controls src=""></video>
	</div>
	<div class="modal-footer">
		<a href="#" class="btn" data-dismiss="modal" aria-hidden="true">{{ _('Close') }}</a>
	</div>
</div>
